import { CacheConsultant } from './CacheConsultant.js';
import { plugins } from './plugins.js';
import { debugLogArray, debugLogObject } from './util/debug.js';
import { _pureGlob, hasNoProductionSuffix, hasProductionSuffix, negate, prependDirToPattern } from './util/glob.js';
import { get, getKeysByValue } from './util/object.js';
import { basename, dirname, join, toPosix } from './util/path.js';
import { getFinalEntryPaths, loadConfigForPlugin } from './util/plugin.js';
import { fromEntryPattern, fromProductionEntryPattern, isEntryPattern, isProductionEntryPattern, } from './util/protocols.js';
const nullConfig = { config: null, entry: null, project: null };
const initEnabledPluginsMap = () => Object.keys(plugins).reduce((enabled, pluginName) => ({ ...enabled, [pluginName]: false }), {});
export class WorkspaceWorker {
    name;
    dir;
    cwd;
    config;
    manifest;
    manifestScriptNames;
    dependencies;
    isProduction;
    isStrict;
    rootIgnore;
    negatedWorkspacePatterns = [];
    enabledPluginsMap = initEnabledPluginsMap();
    enabledPlugins = [];
    enabledPluginsInAncestors;
    cache;
    constructor({ name, dir, cwd, config, manifest, dependencies, isProduction, isStrict, rootIgnore, negatedWorkspacePatterns, enabledPluginsInAncestors, }) {
        this.name = name;
        this.dir = dir;
        this.cwd = cwd;
        this.config = config;
        this.manifest = manifest;
        this.manifestScriptNames = new Set(Object.keys(manifest.scripts ?? {}));
        this.dependencies = dependencies;
        this.isProduction = isProduction;
        this.isStrict = isStrict;
        this.rootIgnore = rootIgnore;
        this.negatedWorkspacePatterns = negatedWorkspacePatterns;
        this.enabledPluginsInAncestors = enabledPluginsInAncestors;
        this.cache = new CacheConsultant(`plugins-${name}`);
    }
    async init() {
        this.enabledPlugins = await this.determineEnabledPlugins();
    }
    async determineEnabledPlugins() {
        const manifest = this.manifest;
        const pluginEntries = Object.entries(plugins);
        for (const [pluginName, plugin] of pluginEntries) {
            if (this.config[pluginName] === false)
                continue;
            if (this.config[pluginName]) {
                this.enabledPluginsMap[pluginName] = true;
                continue;
            }
            const isEnabledInAncestor = this.enabledPluginsInAncestors.includes(pluginName);
            if (isEnabledInAncestor ||
                (await plugin.isEnabled({ cwd: this.dir, manifest, dependencies: this.dependencies, config: this.config }))) {
                this.enabledPluginsMap[pluginName] = true;
            }
        }
        const enabledPlugins = getKeysByValue(this.enabledPluginsMap, true);
        const enabledPluginTitles = enabledPlugins.map(name => plugins[name].title);
        debugLogObject(this.name, 'Enabled plugins', enabledPluginTitles);
        return enabledPlugins;
    }
    getConfigForPlugin(pluginName) {
        const config = this.config[pluginName];
        return typeof config === 'undefined' || typeof config === 'boolean' ? nullConfig : config;
    }
    getEntryFilePatterns() {
        const { entry } = this.config;
        if (entry.length === 0)
            return [];
        const excludeProductionNegations = entry.filter(pattern => !(pattern.startsWith('!') && pattern.endsWith('!')));
        return [excludeProductionNegations, this.negatedWorkspacePatterns].flat();
    }
    getProjectFilePatterns(testFilePatterns) {
        const { project } = this.config;
        if (project.length === 0)
            return [];
        const excludeProductionNegations = project.filter(pattern => !(pattern.startsWith('!') && pattern.endsWith('!')));
        const negatedPluginConfigPatterns = this.getPluginConfigPatterns().map(negate);
        const negatedPluginProjectFilePatterns = this.getPluginProjectFilePatterns().map(negate);
        return [
            excludeProductionNegations,
            negatedPluginConfigPatterns,
            negatedPluginProjectFilePatterns,
            testFilePatterns,
            this.negatedWorkspacePatterns,
        ].flat();
    }
    getPluginProjectFilePatterns() {
        const patterns = [];
        for (const [pluginName, plugin] of Object.entries(plugins)) {
            const pluginConfig = this.getConfigForPlugin(pluginName);
            if (this.enabledPluginsMap[pluginName]) {
                const { entry, project } = pluginConfig;
                patterns.push(...(project ?? entry ?? plugin.project ?? []));
            }
        }
        return [patterns, this.negatedWorkspacePatterns].flat();
    }
    getPluginConfigPatterns() {
        const patterns = [];
        for (const [pluginName, plugin] of Object.entries(plugins)) {
            const pluginConfig = this.getConfigForPlugin(pluginName);
            if (this.enabledPluginsMap[pluginName] && pluginConfig) {
                const { config } = pluginConfig;
                patterns.push(...(config ?? plugin.config ?? []));
            }
        }
        return patterns;
    }
    getProductionEntryFilePatterns(negatedTestFilePatterns) {
        const entry = this.config.entry.filter(hasProductionSuffix);
        if (entry.length === 0)
            return [];
        const negatedEntryFiles = this.config.entry.filter(hasNoProductionSuffix).map(negate);
        return [entry, negatedEntryFiles, negatedTestFilePatterns, this.negatedWorkspacePatterns].flat();
    }
    getProductionProjectFilePatterns(negatedTestFilePatterns) {
        const project = this.config.project;
        if (project.length === 0)
            return this.getProductionEntryFilePatterns(negatedTestFilePatterns);
        const _project = this.config.project.map(pattern => {
            if (!(pattern.endsWith('!') || pattern.startsWith('!')))
                return negate(pattern);
            return pattern;
        });
        const negatedEntryFiles = this.config.entry.filter(hasNoProductionSuffix).map(negate);
        const negatedPluginConfigPatterns = this.getPluginConfigPatterns().map(negate);
        const negatedPluginProjectFilePatterns = this.getPluginProjectFilePatterns().map(negate);
        return [
            _project,
            negatedEntryFiles,
            negatedPluginConfigPatterns,
            negatedPluginProjectFilePatterns,
            negatedTestFilePatterns,
            this.negatedWorkspacePatterns,
        ].flat();
    }
    getConfigurationFilePatterns(pluginName) {
        const plugin = plugins[pluginName];
        const pluginConfig = this.getConfigForPlugin(pluginName);
        if (pluginConfig) {
            return pluginConfig.config ?? plugin.config ?? [];
        }
        return [];
    }
    getIgnorePatterns() {
        return [...this.rootIgnore, ...this.config.ignore.map(pattern => prependDirToPattern(this.name, pattern))];
    }
    async findDependenciesByPlugins() {
        const entryFilePatterns = new Set();
        const productionEntryFilePatterns = new Set();
        const referencedDependencies = new Set();
        const name = this.name;
        const cwd = this.dir;
        const baseOptions = {
            rootCwd: this.cwd,
            cwd,
            manifest: this.manifest,
            manifestScriptNames: this.manifestScriptNames,
            dependencies: this.dependencies,
            isProduction: this.isProduction,
            enabledPlugins: this.enabledPlugins,
        };
        for (const [pluginName, plugin] of Object.entries(plugins)) {
            if (this.enabledPluginsMap[pluginName]) {
                const hasResolveEntryPaths = typeof plugin.resolveEntryPaths === 'function';
                const hasResolveConfig = typeof plugin.resolveConfig === 'function';
                const shouldRunConfigResolver = hasResolveConfig && (!this.isProduction || (this.isProduction && 'production' in plugin));
                const hasResolve = typeof plugin.resolve === 'function';
                const pluginConfig = this.getConfigForPlugin(pluginName);
                if (!pluginConfig)
                    continue;
                const patterns = this.getConfigurationFilePatterns(pluginName);
                const allConfigFilePaths = await _pureGlob({ patterns, cwd, gitignore: false });
                const { packageJsonPath } = plugin;
                const configFilePaths = allConfigFilePaths.filter(filePath => basename(filePath) !== 'package.json' ||
                    (typeof packageJsonPath === 'function'
                        ? packageJsonPath(this.manifest)
                        : get(this.manifest, packageJsonPath ?? pluginName)));
                debugLogArray([name, plugin.title], 'config file paths', configFilePaths);
                const pluginDependencies = new Set();
                const addDependency = (specifier, configFilePath) => {
                    pluginDependencies.add(specifier);
                    if (isEntryPattern(specifier)) {
                        entryFilePatterns.add(fromEntryPattern(specifier));
                    }
                    else if (isProductionEntryPattern(specifier)) {
                        productionEntryFilePatterns.add(fromProductionEntryPattern(specifier));
                    }
                    else if (configFilePath) {
                        referencedDependencies.add([configFilePath, toPosix(specifier)]);
                    }
                };
                const options = { ...baseOptions, config: pluginConfig, configFileDir: cwd, configFileName: '' };
                const configEntryPaths = [];
                for (const configFilePath of configFilePaths) {
                    const opts = { ...options, configFileDir: dirname(configFilePath), configFileName: basename(configFilePath) };
                    if (hasResolveEntryPaths || shouldRunConfigResolver) {
                        const isManifest = basename(configFilePath) === 'package.json';
                        const fd = isManifest ? undefined : this.cache.getFileDescriptor(configFilePath);
                        if (fd?.meta?.data && !fd.changed) {
                            if (fd.meta.data.resolveEntryPaths)
                                for (const id of fd.meta.data.resolveEntryPaths)
                                    configEntryPaths.push(id);
                            if (fd.meta.data.resolveConfig)
                                for (const id of fd.meta.data.resolveConfig)
                                    addDependency(id, configFilePath);
                        }
                        else {
                            const config = await loadConfigForPlugin(configFilePath, plugin, opts, pluginName);
                            const data = {};
                            if (config) {
                                if (hasResolveEntryPaths) {
                                    const dependencies = (await plugin.resolveEntryPaths?.(config, opts)) ?? [];
                                    for (const id of dependencies)
                                        configEntryPaths.push(id);
                                    data.resolveEntryPaths = dependencies;
                                }
                                if (shouldRunConfigResolver) {
                                    const dependencies = (await plugin.resolveConfig?.(config, opts)) ?? [];
                                    for (const id of dependencies)
                                        addDependency(id, configFilePath);
                                    data.resolveConfig = dependencies;
                                }
                                if (!isManifest && fd?.changed && fd.meta)
                                    fd.meta.data = data;
                            }
                        }
                    }
                }
                const finalEntryPaths = getFinalEntryPaths(plugin, options, configEntryPaths);
                for (const id of finalEntryPaths)
                    addDependency(id);
                if (hasResolve) {
                    const dependencies = (await plugin.resolve?.(options)) ?? [];
                    for (const id of dependencies)
                        addDependency(id, join(cwd, 'package.json'));
                }
                debugLogArray([name, plugin.title], 'dependencies', pluginDependencies);
            }
        }
        return {
            entryFilePatterns,
            productionEntryFilePatterns,
            referencedDependencies,
            enabledPlugins: this.enabledPlugins,
        };
    }
    onDispose() {
        this.cache.reconcile();
    }
}
