import { existsSync } from 'node:fs';
import { isBuiltin } from 'node:module';
import { createMatchPath } from 'tsconfig-paths';
import ts from 'typescript';
import { DEFAULT_EXTENSIONS } from '../constants.js';
import { timerify } from '../util/Performance.js';
import { sanitizeSpecifier } from '../util/modules.js';
import { dirname, extname, isAbsolute, isInNodeModules, join, toPosix } from '../util/path.js';
import { resolveSync } from '../util/resolve.js';
import { isDeclarationFileExtension } from './ast-helpers.js';
const resolutionCache = new Map();
const fileExists = (name, containingFile) => {
    const resolvedFileName = isAbsolute(name) ? name : join(dirname(containingFile), name);
    if (existsSync(resolvedFileName)) {
        return {
            resolvedFileName,
            extension: extname(name),
            isExternalLibraryImport: false,
            resolvedUsingTsExtension: false,
        };
    }
};
export function createCustomModuleResolver(compilerOptions, customCompilerExtensions, toSourceFilePath, useCache = true) {
    const customCompilerExtensionsSet = new Set(customCompilerExtensions);
    const extensions = [...DEFAULT_EXTENSIONS, ...customCompilerExtensions];
    function resolveModuleNames(moduleNames, containingFile) {
        return moduleNames.map(moduleName => {
            if (!useCache)
                return resolveModuleName(moduleName, containingFile);
            const key = moduleName.startsWith('.')
                ? join(dirname(containingFile), moduleName)
                : `${containingFile}:${moduleName}`;
            if (resolutionCache.has(key))
                return resolutionCache.get(key);
            const resolvedModule = resolveModuleName(moduleName, containingFile);
            if (resolvedModule)
                resolutionCache.set(key, resolvedModule);
            return resolvedModule;
        });
    }
    const tsMatchPath = createMatchPath(compilerOptions.baseUrl ?? '/', compilerOptions.paths || {});
    function resolveModuleName(name, containingFile) {
        const sanitizedSpecifier = sanitizeSpecifier(name);
        if (isBuiltin(sanitizedSpecifier) || isInNodeModules(name))
            return undefined;
        const resolvedFileName = resolveSync(sanitizedSpecifier, containingFile, extensions);
        if (resolvedFileName) {
            const ext = extname(resolvedFileName);
            if (!customCompilerExtensionsSet.has(ext)) {
                const srcFilePath = toSourceFilePath(resolvedFileName);
                if (srcFilePath) {
                    return {
                        resolvedFileName: srcFilePath,
                        extension: extname(srcFilePath),
                        isExternalLibraryImport: false,
                        resolvedUsingTsExtension: false,
                    };
                }
            }
            return {
                resolvedFileName,
                extension: customCompilerExtensionsSet.has(ext) ? ts.Extension.Js : ext,
                isExternalLibraryImport: isInNodeModules(resolvedFileName),
                resolvedUsingTsExtension: false,
            };
        }
        const pathMappedFileName = tsMatchPath(sanitizedSpecifier, undefined, undefined, []);
        if (pathMappedFileName) {
            const ext = extname(pathMappedFileName);
            return {
                resolvedFileName: toPosix(pathMappedFileName),
                extension: customCompilerExtensionsSet.has(ext) ? ts.Extension.Js : ext,
                isExternalLibraryImport: false,
            };
        }
        const tsResolvedModule = ts.resolveModuleName(sanitizedSpecifier, containingFile, compilerOptions, ts.sys).resolvedModule;
        if (tsResolvedModule) {
            if (isDeclarationFileExtension(tsResolvedModule.extension)) {
                const srcFilePath = toSourceFilePath(tsResolvedModule.resolvedFileName);
                if (srcFilePath) {
                    return {
                        resolvedFileName: srcFilePath,
                        extension: extname(srcFilePath),
                        isExternalLibraryImport: false,
                        resolvedUsingTsExtension: false,
                    };
                }
            }
            return tsResolvedModule;
        }
        const module = fileExists(sanitizedSpecifier, containingFile);
        if (module)
            return module;
        return undefined;
    }
    return timerify(resolveModuleNames);
}
